/* 
 * (C) 2003 Ulrich Hecht <uli@emulinks.de>
 * if2a patch by D. Gauchard <deyv@free.fr>
 * F2A Ultra support by Vincent Rubiolo <vincent.rubiolo@free.fr>
 * Licensed under the terms of the GNU Public License version 2
 */
 
// Cart/ROM functions

#include "binware.h"
#include "f2acarts.h"
#include "f2aultra.h"
#include "f2autils.h"
#include "f2aio.h"
#include "cartmap.h"

int f2a_reduce_forced = 0;
int f2a_reduce_allowed = 1;
int f2a_cartsize_mb = -1;		/* autodetect, see f2a_get_type() */
int f2a_write_block_size_log2 = -1;	/* autodetect, see f2a_get_type() */
int f2a_rom_block_size_log2 = -1;	/* autodetect, see f2a_get_type() */
int f2a_thorough_comp = 0;
int f2a_use_loader = 1;
int f2a_correct_header_allowed = 1;
int f2a_force_burn = 0;

int select_loader (enum cart_type_e cart_type, char* loader_file)
{
	int result = -1;

	if (loader.size == 0)
	{
		switch (cart_type)
		{
			case CART_TYPE_PRO:
				result = select_loader_pro(loader_file);
				break;
			case CART_TYPE_ULTRA:
				result = select_loader_ultra(loader_file);
				break;
			case CART_TYPE_TURBO:
				printerr("looking for TURBO cart testers\n");
				result = select_loader_pro(loader_file);
				break;
			case CART_TYPE_UNDEF:
				printerr("Undefined cart type, please use -t\n");
				result = -1;
				break;
		}
		return result;
	}
	else
		return 0;
}

int trim (char* rom, int size)
{
	char last;
	int trimmed_size;
	
	if (f2a_reduce_allowed)
	{
		// trim to get interesting size
		last = rom[size - 1];
		for (trimmed_size = size - 2; trimmed_size > 0 && rom[trimmed_size] == last; trimmed_size--);
		trimmed_size += 2;
	}
	else
		trimmed_size = size;
	
	return trimmed_size;
}

// this is pogoshell header detection test:
//if( ((ptr[3] & 0xFE) == 0xEA) && (ptr[0xb2] == 0x96))

int has_good_header (unsigned char* rom)
{
	/* 
	 * FIXME
	 * Weak implementation : no check on required checksum fields
	 */
	return    memcmp(binware_standard_header[0].data+4, rom+4, HEADERSIZE_B - 4) == 0
	       && ((rom[3] & 0xfe) == 0xea)
	       && (rom[0xb2] == 0x96)
	       ? 1: -1;

}

void correct_header (unsigned char* rom, unsigned char* name, int force_correction, int force_name)
{
	unsigned char romname[13];
	unsigned char romcode[5];
	unsigned char* shortname;
	
	memset(romname, 32, 12);
	memset(romcode, 32, 12);
	romname[12] = romcode[4] = 0;

	shortname = filename2romname(name);
	memcpy(romname, shortname, MIN(strlen(shortname) + 1, 12));
	memcpy(romcode, shortname, MIN(strlen(shortname), 4));
	
	if (force_name || !isalnum(rom[0xa0]))
		memcpy(&rom[0xa0], romname, 12);
	if (!isalnum(rom[0xac]))
		memcpy(&rom[0xac], romcode, 4);
	//if (!isalnum(rom[0xb0]))
	//	rom[0xb0] = '0';
	
	if (force_correction || has_good_header(rom) < 0)
	{
		if (!force_correction)
			printerr("No standard header found in '%s', rewriting it\n", romname);
		memcpy(rom, binware_standard_header[0].data, HEADERSIZE_B);
		// from pogo
		rom[3] = (rom[3] & ~0xfe) | 0xea;
		rom[0xb2] = 0x96;
	}
}

void display_map (unsigned char* rom)
		{
	int i;
	
	print("Game Title:	");
	for (i = 0xa0; i < 0xac; i++)
		print("%c", ASCII(rom[i]));
	print("\nGame Code:	AGB-");
	for (i = 0xac; i < 0xb0; i++)
		print("%c", ASCII(rom[i]));
	print("\nMaker Code:	");
	for (i = 0xb0; i < 0xb2; i++)
		print("%c", ASCII(rom[i]));
	print("\n");
	//print("\n0x96:		0x%02x\n", rom[0xb2]);
}

unsigned char* romname (unsigned char* rom)
{
	int len;
	static char name [13];
	
	for (len = 0; len < 12; len++)
		name[len] = ASCII(rom[0xa0 + len]);
	for (len--; len>=0 && (name[len]==32 || name[len]=='.'); len--);
	name[len+1] = 0;
	return name;
}

// remove directory part and extension, limit to 12 chars
unsigned char* filename2romname (unsigned char* filename)
{
	int i, len;
	static char name [13];
	
	len = 0;
	for (i = strlen(filename) - 1; i >= 0 && filename[i] != '/' && filename[i] != '\\'; i--)
		if (filename[i] == '.')
			len = 0;
		else
			len++;
	if (len > 12)
		len = 12;
	strncpy(name, &filename[i + 1], 12);
	name[len] = 0;

	return name;
}

int old_display_cart_map (void)
{
	int i;
	unsigned char rom[LOADUNIT];
	
	if (f2a_sim > 1)
		return 0;

	for (i = GBA_ROM; i < GBA_ROM + f2a_max_rom_size; i += f2a_rom_block_size)
	{
		print("Trying address 0x%x...\r", i);
		if (f2a_readmem(rom, i, LOADUNIT) == -1)
			return -1;
		if (has_good_header(rom) > -1)
		{
			print("\n");
			display_map(rom);
		}
	}
	print("\n");

	return 0;
}

void auto_readandsave_rom (int numfiles, char** files)
{
	int i;
	int indexfile;
	int nextblockloaded;
	char* name;
	char autoname[64];
	unsigned char rom[f2a_rom_block_size];
	
	assert(f2a_rom_block_size >= LOADUNIT);
	
	if (f2a_sim > 1)
		return;

	indexfile = 0;
	i = GBA_ROM;
	f2a_readmem(rom, i, LOADUNIT);
	do
	{
		nextblockloaded = 0;
		
		print("Trying address 0x%x...\r", i);
		if (has_good_header(rom) != -1)
		{
			/* Prepare filename */
			name = romname(rom);
			strcpy(autoname, GBA_HEADNAME);
			strcat(autoname, name);
			strcat(autoname, GBA_EXTNAME);
			
			print("\n");
			print("Detected rom '%s'\n", name);
			
			/* If no arg is given, autoname; otherwise, use specified name */
			if (indexfile >= numfiles)
				name = autoname;
			else
				name = files[indexfile];
			indexfile++;
			
			if (strcmp(name, "@") == 0)
				name = autoname;
			
			if (strcmp(name, ".") == 0) /* user don't want this rom */
			{
				if (f2a_verbose)
					print("Skipping on user request\n");
			}
			else
			{
				FILE* f;
				if (f2a_verbose)
					print("Using name file name '%s'\n", name);
				if ((f = fopen(name, "wb")) == NULL)
					printerrno("fopen(%s)", name);
				else
				{
					if (f2a_readmem(rom, i, f2a_rom_block_size) < 0)
						return;
					do /* read data up to next ROM/valid header */
					{
						if (fwrite(rom, f2a_rom_block_size, 1, f) != 1)
						{
							if (ferror(f))
								printerrno("write(%s)", name);
							else
								printerr("could not write to file %s\n", name);
							break;
						}
						if ((i+=f2a_rom_block_size) < GBA_ROM + f2a_max_rom_size && f2a_readmem(rom, i, f2a_rom_block_size) < 0)
							return;
					} while (i < GBA_ROM + f2a_max_rom_size && has_good_header(rom) == -1);
					fclose(f);
					nextblockloaded = 1;
				}
			}
		}
		
		if (!nextblockloaded) // continue to look for ROM/valid header
		{
			if ((i+=f2a_rom_block_size) < GBA_ROM + f2a_max_rom_size && f2a_readmem(rom, i, LOADUNIT) < 0)
				return;
		}
	} while (i < GBA_ROM + f2a_max_rom_size);
	
	print("\n");
}

void display_memory_map (unsigned char* rom, int size)
{
	int i;

	for (i = 0; i < size; i += f2a_rom_block_size)
	{
		print("Trying address 0x%x...\r", i);
		if (has_good_header(&rom[i]) > -1)
		{
			print("\n");
			display_map(&rom[i]);
		}
	}
}

int is_same (int offset, int size, unsigned char* mem)
{
	int ret, i;
	unsigned char rom [LOADUNIT];
	
	if (f2a_sim > 1)
		return -1;

	for (i = offset; i < offset + size; i += LOADUNIT)
	{
		// check beginning
		f2a_readmem(rom, GBA_ROM + i, LOADUNIT);
		ret = memcmp(rom, &mem[i], LOADUNIT);
		if (f2a_verbose > 1 && ret != 0)
			print("(offset 0x%x-0x%x BAD)\n", i, i + LOADUNIT - 1);
		if (ret != 0)
			return -1;
	}
	return 1;
}

int has_same_data (int offset, int size, unsigned char* mem)
{
	if (f2a_thorough_comp)
	{
		int i;
		for (i = offset; i < offset + size; i += f2a_rom_block_size)
			if (is_same(i, f2a_rom_block_size, mem) == -1)
				return -1;
	}
	else
	{
		// check beginnning
		if (is_same(offset, LOADUNIT, mem) == -1)
			return -1;

		// roughly check end
		size &= ~(LOADUNIT-1);
		if (is_same(offset + size - LOADUNIT, LOADUNIT, mem) == -1)
			return -1;
	}
	return 1;
}

int convsize (char* size)
{
	char* unit = size;
	while (isdigit(*unit))
		unit++;
	if (strlen(unit) == 2)
	{
		int byte;
		switch (tolower(unit[0]))
		{
		case 'k':	byte = 1<<10; break;
		case 'm':	byte = 1<<20; break;
		default:	byte = -8;
		}
		switch (unit[1])
		{
		case 'b':	byte >>= 3; break;
		case 'B':	break;
		default:	byte = -1;
		}
		if (byte > 0)
		{
			*unit = 0; // THIS IS BAD, we should use a const char* as argument
			return atoi(size) * byte;
		}
	}
	printerr("size '%s' not recognized\n", size);
	f2a_exit(1);
	return -1;
}

int convf2ablock (char* block, int* offset, int* size )
{
	int len = strlen(block);

	if (len > 3)
		return -1;

	if (strcmp(block, "all") == 0 || block[0])
	{
		*offset = 0;
		*size = 256 << 10;
	}
	else
	{
		*offset = (block[0] - '1') << 16;
		if (block[1])
		{
			*offset += (tolower(block[1]) - 'a') << 15;
			
			if (block[2])
			{
				*offset += (block[2] - '1') << 13;
				*size = 1 << 13;
			}
			else
				*size = 1 << 15;
		}
		else
			*size = 1 << 16;
	}
	if (f2a_verbose && block[0])
		print("Block %s is SRAM is (offset=0x%x/%ibytes size=0x%x/%ibytes)\n",
			block, *offset, *offset, *size, *size);
	return 0;
}

//////////////////////////////////////////////////////////////////////////
// check whole size

#define STAT							\
		if (index == -1)				\
			st.st_size = loader.size;		\
		else if (stat(files[index], &st) == -1)		\
		{						\
			perror(files[index]);			\
			return -1;				\
		} 

int get_wholesize (int numfiles, char** files)
{
	int index;
	struct stat st;
	int wholesize = 0;
	int roundedfilesize;
	
	// stat every file
	for (index = f2a_use_loader? -1: 0; index < numfiles; index++)
	{
		if (f2a_verbose)
			print("Stat file: %s\n", index==-1? "Loader": files[index]);

		STAT

		roundedfilesize = st.st_size;
		adjust_rom_size(&roundedfilesize); // round on f2a_rom_block_size
		wholesize += roundedfilesize;
	}
	
/*
	if (wholesize > f2a_max_rom_size + loader.size)
	{
		printerr("You have not enough memory !\n");
		f2a_exit(1);
	}
*/
	if (wholesize > f2a_max_rom_size)
		// We hope that we will be able to reduce overall content by loader.size
		wholesize = f2a_max_rom_size;
	
	adjust_write_size(&wholesize); // round on f2a_write_block_size
	return wholesize;
}

//////////////////////////////////////////////////////////////////////////
// this one is simple, it writes one file to sram

int f2apro_loadandwrite_sram (char* file, int offset, int size)
{
	int ret;
	char* sram;
	
	if ((sram = prepare_loadandwrite_sram (file, offset, size)) == NULL)
		return -1;
	ret = f2a_writemem(sram, GBA_SRAM, offset, size, size, 0);
	free(sram);
	return ret;
}

char* prepare_loadandwrite_sram (char* file, int offset, int size)
{
	FILE* sramf;
	struct stat st;
	char* sram = NULL;
	
	if (stat(file, &st) == -1)
	{
		perror(file);
		return NULL;
	}

	// some checks
	if (st.st_size > size)
	{
		printerr("File too big (0x%x bytes) to fit in 0x%x bytes\n",
			(int)st.st_size, size);
		return NULL;
	}

	// allocate memory
	if ((sram = (unsigned char*)malloc(size)) == NULL)
	{
		printerr("Cannot allocate 0x%x/%i bytes for sram management - ", size, size);
		perror("malloc");
		return NULL;
	}
	memset(sram, 0xff, size);

	// load files into "allocated rom"
	if (f2a_verbose)
		print("Loading file: %s (size=0x%x)\n", file, (int)st.st_size);

	if ((sramf = fopen(file, "rb")) == NULL)
	{
		printerrno("fopen(%s)", file);
		free(sram);
		return NULL;
	}
	if (fread(sram, size, 1, sramf) != 1)
	{
		if (ferror(sramf))
			printerrno("fread(%s)", file);
		else
			printerr("failed to read from file %s", file);
		free(sram);
		return NULL;
	}
	fclose(sramf);
	
	return sram;
}


//////////////////////////////////////////////////////////////////////////
// automatic check cart and burn if necessary

int auto_loadandburn_rom (enum cart_type_e cart_type, int clean_rom, int numfiles, char** files)
{
	typedef struct
	{
		int	offset;
		int	size;
	} chunk_s;

	int			burnstart		= -1;	// start address of current chunk
	int			wholesize		= get_wholesize(numfiles, files);
	cart_map_locator_s*	locator			= NULL;
	cart_map_s*		cart_map		= NULL;
	int			cart_map_entries	= 0;
	int			chunks_number		= numfiles + 1;
	int			chunks_used		= 0;
	int			whole_loader_size	= 0;
	int			reduce			= 0;

	int			index;
	FILE*			ROMf;
	struct stat		st;
	int			loadedsize;
	int			roundedfilesize;
	char*			image;				// image that will contain all cart contents
	chunk_s			chunks [chunks_number];		// chunks to burn array descriptor
	content_desc		desc;
	
	if (!wholesize)
		return -1;
	
	// allocate image memory
	if ((image = (unsigned char*)malloc(wholesize)) == NULL)
	{
		printerrno("Cannot allocate 0x%x/%i bytes for cart management - malloc", wholesize, wholesize);
		return -1;
	}
	memset(image, 0xff, wholesize);
	
	// load
	loadedsize = 0;
	for (index = (f2a_use_loader? -1: 0); index < numfiles; index++)
	{
		print("\n");
		
		STAT
		
		if (index == -1) // manage loader, fit cart map
		{
			int i, real_loader_size;
			char last = loader.data[loader.size - 1];

			assert(loader.size > 0);

			roundedfilesize = loader.size;
			adjust_rom_size(&roundedfilesize);
			// copy f2aloader into image
			memcpy(image, loader.data, loader.size);
			// pad chunk with last loader byte
			memset(image + loader.size, last, roundedfilesize - loader.size);

			// try to find a location for the cart map
			for (i = roundedfilesize - 2; image[i] == last; i--);
			real_loader_size = i + 2;
			assert(real_loader_size <= loader.size);
			
			// check whether cart map can fit 
			do
			{
				int spare_size = roundedfilesize - real_loader_size;
				cart_map_entries = spare_size > sizeof(cart_map_locator_s)? (spare_size - sizeof(cart_map_locator_s)) / sizeof(cart_map_s): 0;
				if (f2a_verbose)
					print("Spare size for cart map at rounded loader's end is %i bytes, which gives room for %i roms.\n", spare_size, cart_map_entries);
				if (cart_map_entries >= MAP_MINIMUM_ENTRIES)
				{
					int cart_map_locator = roundedfilesize - sizeof(cart_map_locator_s);
					locator = (cart_map_locator_s*)&image[cart_map_locator];
					locator->location = hton32(roundedfilesize - sizeof(cart_map_locator_s) - cart_map_entries * sizeof(cart_map_s));
					locator->number_of_entries = hton16(cart_map_entries);
					locator->magic = hton32(MAP_MAGIC);
					cart_map = (cart_map_s*)&image[ntoh32(locator->location)];
					if (f2a_verbose)
						print("Rom map table is at address 0x%x\nRom map locator is at address 0x%x (end of loader)\n",
							GBA_ROM + ntoh32(locator->location),
							GBA_ROM + cart_map_locator);
				}
				else
				{
					int oldfilesize = roundedfilesize;
					roundedfilesize++;
					adjust_rom_size(&roundedfilesize);
					// pad chunk with last loader byte
					memset(image + oldfilesize, last, roundedfilesize - oldfilesize);
					print("Increasing loader size to fit cart map by 0x%x/%i bytes\n", roundedfilesize - oldfilesize, roundedfilesize - oldfilesize);
				}
			} while (cart_map_entries < MAP_MINIMUM_ENTRIES);
			if (f2a_verbose)
				print("\n");

			print("Loader %s is:\n", loader.name);
			whole_loader_size = roundedfilesize;
		}
		else
		{
			print("File %s at address 0x%x is:\n", files[index], GBA_ROM + loadedsize);

			roundedfilesize = st.st_size;
			adjust_rom_size(&roundedfilesize);
			
			if (loadedsize + st.st_size > f2a_max_rom_size)
			{
				printerr("! WARNING ! Cart size exceeded, truncating file (size 0x%x -> 0x%x bytes) !\n", (int)st.st_size, f2a_max_rom_size - loadedsize);
				if ((st.st_size = f2a_max_rom_size - loadedsize) == 0)
					break;
				roundedfilesize = st.st_size;
				adjust_rom_size(&roundedfilesize);
			}

			// load ROM into image file
			if ((ROMf = fopen(files[index], "rb")) == NULL)
			{
				if (ferror(ROMf))
					printerrno("fopen(%s)", files[index]);
				else
					printerr("Could not open file %s for reading\n", files[index]);
				return -1;
			}
			if (fread(&image[loadedsize], st.st_size, 1, ROMf) != 1)
			{
				if (ferror(ROMf))
					printerrno("fread(%s)", files[index]);
				else
					printerr("Could not read file %s\n", files[index]);
				return -1;
			}
			fclose(ROMf);
			
			// try to reduce rom size
			if (f2a_reduce_allowed || f2a_reduce_forced)
			{
				int i, min_limit;
				// last bytes are probably padding
				char last = image[loadedsize + st.st_size - 1];

				// f2a_reduce_forced tries to mad-pad everything
				// !f2a_reduce_forced tries to fit loader only
				min_limit = f2a_reduce_forced? 0: MAX(0, (st.st_size - whole_loader_size - 1));
				// truncate as long as there is padding
				for (i = st.st_size - 2; i >= min_limit && image[loadedsize + i] == last; i--);
				roundedfilesize = i + 2;
				adjust_rom_size(&roundedfilesize);

				if (roundedfilesize < st.st_size)
				{
					print("(reducing ROM size from %ikB to %ikB", 
						((int)st.st_size + 1023) >> 10, 
						(roundedfilesize + 1023) >> 10);
					if (   (reduce += ((int)st.st_size - roundedfilesize)) > whole_loader_size
					    && !f2a_reduce_forced)
					{
						print(" - not trying to further reduce ROM size");
						f2a_reduce_allowed = 0;
					}
					print(")\n");
				}
			}
		}

		// check file against real ROM
		if (f2a_correct_header_allowed && index != -1)
			correct_header(&image[loadedsize], files[index], 0, 0);
		display_map(&image[loadedsize]);
		
		// fill the cart locator
		if (locator)
		{
			if (index >= cart_map_entries)
			{
				printerr("Warning: not enough room in cart locator to stock further rom entries !\n");
				locator = NULL;
			}
			else
			{
				strncpy(cart_map[index].name, romname(&image[loadedsize]), MAP_NAMELEN);
				cart_map[index].offset = hton32(loadedsize);
				cart_map[index].size = hton32(roundedfilesize);
			}
		}
		
		if (!f2a_force_burn && has_same_data(loadedsize, roundedfilesize, image) >= 0)
		{
			print("No need to burn it!\n");
			// but it's time to burn the previous ones if they changed
			if (burnstart >= 0)
			{
				//XXXFIXME remove all this commentary when ACKed
				//	we don't burn anymore while loading, but record the location
				//	to burn into the chunks array because the map locator
				//	needs to be burned after all files are loaded, although it is
				//	located itself in the beginning of the cart
				//if (f2a_burn(GBA_ROM, image, burnstart, loadedsize - burnstart) < 0)
				//	return -1;
				
				// register chunk
				assert(chunks_used < chunks_number);
				chunks[chunks_used].offset = burnstart;
				chunks[chunks_used].size = loadedsize - burnstart;
				chunks_used++;
				
				// next chunk will be a new separate one
				burnstart = -1;
			}
		}
		else
		{
			print("Not the same !\n");
			// and we remember start address
			if (burnstart < 0)
				burnstart = loadedsize;
		}

		loadedsize += roundedfilesize;
	}

	// mark end of map
	if (locator && index < cart_map_entries)
	{
		cart_map[index].name[0] = 0;
		cart_map[index].offset = hton32(0);
		cart_map[index].size = hton32(0);
	}

	// burn the last chunk (XXXFIXMEremove this commentary)
	//if (burnstart >= 0 && f2a_burn(GBA_ROM, image, burnstart, loadedsize - burnstart))
	//	return -1;

	// register the last chunk
	if (burnstart >= 0)
	{
		assert(chunks_used < chunks_number);
		chunks[chunks_used].offset = burnstart;
		chunks[chunks_used].size = loadedsize - burnstart;
		chunks_used++;
	}
	
	print("\n");
	// now burn all the recorded chunks
	for (index = 0; index < chunks_used; index++)
		if (f2a_burn(GBA_ROM, 0, image, chunks[index].offset, chunks[index].size) < 0)
			return -1;

        /* 
	 * Temporary write support for Ultra carts : we wipe out the descriptor and rely on
	 * CIZ to regenerate it
	 */
	if (cart_type == CART_TYPE_ULTRA)
	{
		print("Wiping out content descriptor. You will need to launch CIZ to regenerate it.\n");
		memset((char*) &desc, 0x99, sizeof(desc));
		f2a_writemem((unsigned char*) &desc, F2AU_CD_BASE, 0, SIZE_64K, f2a_write_block_size, 0);
	}
	
	print("Remaining size: %iMbits = %iMBytes = %iKBytes\n",
		(f2a_max_rom_size - loadedsize) / 1024 / 1024 * 8,
		(f2a_max_rom_size - loadedsize) / 1024 / 1024,
		(f2a_max_rom_size - loadedsize) / 1024);

	if (clean_rom)
	{
		//int cleanblocksize = f2a_write_block_size; // too big, unnecessary
		//int cleanblocksize = HEADERSIZE_B; // don't work (?)
		int cleanblocksize = f2a_rom_block_size; // 32Kb

		// we write cleanblocksize bytes every f2a_write_block_size, I
		// suppose that each write cleans f2a_write_block_size
		// (>=cleanblocksize) bytes. to be verified...
		char empty[cleanblocksize];

		int dummy = 0;
		print("\nCleaning remaining ROM...\n");
		memset(empty, 0xff, cleanblocksize);
		burnstart = loadedsize + f2a_write_block_size - 1;
		adjust_burn_addresses(&burnstart, &dummy);
		for (; burnstart < f2a_max_rom_size; burnstart += f2a_write_block_size)
		{
			print("Cleaning from 0x%x to 0x%x\n", GBA_ROM + burnstart, 
				GBA_ROM + burnstart + cleanblocksize);
			f2a_writemem(empty, GBA_ROM, burnstart, cleanblocksize, cleanblocksize, 0);
			//break; no break: need to clear everything left
		}
		print("\n");
	}
	
	free(image);
	return 0;
}

// burned addresses have to be on f2a_write_block_size multiples boundaries
void adjust_burn_addresses (int* offset, int* size)
{
	// adjust offset so that it's a multiple of block
	// adjust size so that it's a multiple of block
	*size += *offset;
	*offset &= ~(f2a_write_block_size - 1);
	*size -= *offset;
	if (*size & (f2a_write_block_size - 1))
		*size += f2a_write_block_size;
	*size &= ~(f2a_write_block_size - 1);
}

// burned rom offset and offset+size have to be on f2a_rom_block_size multiples boundaries
void adjust_rom_addresses (int* offset, int* size)
{
	// adjust offset so that it's a multiple of block
	// adjust size so that it's a multiple of block
	// (this is needed if we want to rewrite a rom between two others)
	*size += *offset;
	*offset &= ~(f2a_rom_block_size - 1);
	*size -= *offset;
	if (*size & (f2a_rom_block_size - 1))
		*size += f2a_rom_block_size;
	*size &= ~(f2a_rom_block_size - 1);
}

// load addresses have to be on f2a_rom_block_size multiples boundaries
void adjust_load_addresses (int* offset, int* size)
{
	// adjust offset so that it's a multiple of block
	// adjust size so that it's a multiple of block
	*size += *offset;
	*offset &= ~(LOADUNIT - 1);
	*size -= *offset;
	if (*size & (LOADUNIT - 1))
		*size += LOADUNIT;
	*size &= ~(LOADUNIT - 1);
}

void adjust_write_size (int* size)
{
	if (*size & (f2a_write_block_size - 1))
		*size += f2a_write_block_size;
	*size &= ~(f2a_write_block_size - 1);
}

void adjust_rom_size (int* size)
{
	if (*size & (f2a_rom_block_size - 1))
		*size += f2a_rom_block_size;
	*size &= ~(f2a_rom_block_size - 1);
}
